package com.loit.getway.filter;

import com.loit.common.json.AjaxJson;
import com.loit.common.spring.timeloit.utils.LogUtil;
import com.loit.common.spring.timeloit.utils.LoitStatusCode;
import com.loit.getway.PermissionService;
import com.loit.getway.config.*;
import com.loit.getway.constant.ClientLoitStatusCode;
import com.loit.getway.constant.CommonConstant;
import com.loit.getway.json.ClientAjaxJson;
import com.loit.getway.utils.ClientUserUtil;
import com.loit.getway.utils.GateWayLoitLog;
import com.loit.getway.utils.RedisUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.cloud.gateway.filter.GatewayFilterChain;
import org.springframework.cloud.gateway.filter.GlobalFilter;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.Ordered;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.http.HttpCookie;
import org.springframework.http.MediaType;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.util.AntPathMatcher;
import org.springframework.util.StringUtils;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Calendar;
import java.util.List;

/**
 * @author wbh
 * @creation 2020-03-13 10:52
 */
@Configuration
public class GatewayDataWorksFilter implements GlobalFilter, Ordered {
	protected Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    private AuthSkipUrlsProperties authSkipUrlsProperties;
    @Autowired
    private GateWayLoitLog log;
    @Autowired
    private ClientUserUtil clientUserUtil;

    private AntPathMatcher ant = new AntPathMatcher();
    @Autowired
    GateWayRoutesYml gateWayRoutesYml;
    @Autowired
    GateWayConfig gateWayConfig;
    @Autowired
    private PermissionService permissionService;


    /**
     * 失败的请求会格式 response.getHeaders().setContentType(MediaType.APPLICATION_JSON_UTF8);
     * 成功的请求会通过各个微服务自行转发的json格式
     *
     * @param exchange
     * @param chain
     * @return
     */
    @Override
    public Mono<Void> filter(ServerWebExchange exchange, GatewayFilterChain chain) {
        GatewayContext context = new GatewayContext();
        ServerHttpRequest request = exchange.getRequest();
        ServerHttpResponse response = exchange.getResponse();
		long startTime=System.currentTimeMillis();
        logger.debug(LogUtil.print("0.请求路径="+ request.getURI().getPath()));
        if("/error".equals(request.getURI().getPath())){
            return chain.filter(exchange);
        }

        //获取网关地址端口,用于在服务中获取来源的网关地址
//        String gateWayUrl = request.getURI().getHost();

        String accessToken = null;

        // 获取CommonConstant.TOKEN_NAME, 从cookie中获取, 如果没有则从请求头中获取
        HttpCookie httpCookie = request.getCookies().getFirst(CommonConstant.TOKEN_NAME);
        if(null == httpCookie || StringUtils.isEmpty(httpCookie)){
            accessToken = request.getHeaders().getFirst(CommonConstant.TOKEN_NAME);
            logger.debug(LogUtil.print("1.从请求头={}获取令牌={}"), CommonConstant.TOKEN_NAME, accessToken);
        }else {
        	accessToken = httpCookie.getValue();
            logger.debug(LogUtil.print("1.从cookie={}获取令牌={}"), CommonConstant.TOKEN_NAME, accessToken);
        }

        // 设置单点登录转发路径
//        setCasServerUrl(request);
        /*if (blackServersCheck(context, exchange)) {//服务黑名单（停止指定服务的接口调用）
        	return get903Code(response);
        }*/


        if (isNoLoign()||whiteListCheck(context, exchange)) {//放行的url请求
            logger.debug(LogUtil.print("2.进入放行路径流程,isNoLoign()={},whiteListCheck()={}"), isNoLoign(), whiteListCheck(context, exchange));
            String userId = authToken(context, accessToken, request);
            ServerHttpRequest mutateReq = null;
            if(userId!=null){
            	mutateReq = request.mutate().header(CommonConstant.AUTHORIZATION, userId).build();
            }else {
            	logger.debug(LogUtil.print("userId为空！！ "));
            	mutateReq = request.mutate().build();
            }
            ServerWebExchange mutableExchange = exchange.mutate().request(mutateReq).build();

            // 登出
            if (userId != null && mutateReq.getURI().toString().contains("/cas/logout")) {
                RedisUtil.zRemove(CommonConstant.ONLINE_USERS, userId);
            }
            //如果用户为空并且是非放行url时，表示未登录，返回901
//            if(userId == null && !whiteListCheck(context,exchange)){
//                return get901Code(response, request);
//            }
    		long endTime=System.currentTimeMillis();
    		float excTime=(float)(endTime-startTime);
    		if(excTime>50)
    			log.info("GatewayDataWorksFilter time consuming:" + excTime +"ms");
            return chain.filter(mutableExchange);
        } else {
            logger.debug(LogUtil.print("2.进入不放行路径鉴权流程"));
            // 令牌不存在,无权限访问该资源
            if(null == accessToken || StringUtils.isEmpty(accessToken)){
            	return get901Code(response, request);
            }
        	//token验证
            String userId = authToken(context, accessToken, request);
            if (!context.isDoNext()) {
                // 统计在线用户功能：用户未登录，移除
                RedisUtil.zRemove(CommonConstant.ONLINE_USERS, userId);

            	return get901Code(response, request);
            }
            ServerHttpRequest mutateReq = null;
            if(userId!=null){
            	mutateReq = request.mutate().header(CommonConstant.AUTHORIZATION, userId).build();
            }else {
                logger.debug(LogUtil.print("userId为空！！ "));
            	mutateReq = request.mutate().build();
            }
            ServerWebExchange mutableExchange = exchange.mutate().request(mutateReq).build();

            // 用户在线统计
            this.setOnlineUser(accessToken, userId, mutateReq);

            //资源鉴权
            if(!isNoResourceAuth()){
                //调用门户资源操作验证用户是否有资源操作权限
                checkPermission(context,userId,request);
                if (!context.isDoNext()) {
                    return get903Code(response);
                }
            }

    		long endTime=System.currentTimeMillis();
    		float excTime=(float)(endTime-startTime);
    		if(excTime>50)
    			log.info("GatewayDataWorksFilter time consuming:" + excTime +"ms");
            return chain.filter(mutableExchange);
        }
    }

    /**
     * 用户在线统计设计
     * @param accessToken
     * @param userId
     * @param mutateReq
     */
    private void setOnlineUser(String accessToken, String userId, ServerHttpRequest mutateReq) {
        // 统计在线用户功能：更新时间
        if (accessToken != null && userId != null && !mutateReq.getURI().toString().contains("/cas/login")) {
            Calendar now = Calendar.getInstance();
            now.set(Calendar.SECOND, now.get(Calendar.SECOND) + gateWayConfig.getKillSeconds());
            RedisUtil.zAdd(CommonConstant.ONLINE_USERS, userId, now.getTimeInMillis());
        }
    }

    /**
     * 判断是否免登录
     * @return
     */
	public boolean isNoLoign() {
    	if(!gateWayConfig.getLoginAuth())
    		return true;
    	return false;
    }
    /**
     * @desc 判断是否资源操作免鉴权
     * @return
     * @author yangwenbin
     * @date
     */
    public boolean isNoResourceAuth(){
        if(!gateWayConfig.getResourceAuth())
            return true;
        return false;
    }
    
    /**
     * 获取门户的服务转发路径
     * @return
     */
    public String getPortalLogoutName() {
        List<GateWayService> list = gateWayRoutesYml.getRoutes();
        for (GateWayService gateWayService : list) {
			if(gateWayService.getLogoutSign()!=null) {
				String path = gateWayService.getPredicates().get(0).split("=")[1];
				return path.substring(0, path.indexOf("/*"));
			}
		}
        return null;
    }

    /**
     * 获取Cas单点登录的服务转发路径
     * @return
     */
    public String getCasServerUrlName() {
        List<GateWayService> list = gateWayRoutesYml.getRoutes();
        for (GateWayService gateWayService : list) {
            if(gateWayService.getCasServerSign()!=null) {
                String path = gateWayService.getPredicates().get(0).split("=")[1];
                return path.substring(0, path.indexOf("/*"));
            }
        }
        return null;
    }
    /**
     * @desc 设置单点登录的转发路径
     * @param request
     * @return
     * @author yangwenbin
     * @date
     */
    /*private void setCasServerUrl(ServerHttpRequest request){
        String caseServerUrl = getCasServerUrlName();
        if(!StringUtils.isEmpty(caseServerUrl)){
            URI uri = request.getURI();
            String hostIp = uri.toString().substring(0, uri.toString().indexOf(uri.getPath()));
            gateWayConfig.setServerUrl(hostIp+caseServerUrl);
        }
    }*/
    /**
     * 验证登录
     * @param context
     * @param token 
     * @param request
     */
    private String authToken(GatewayContext context, String token, ServerHttpRequest request) {

        if(noUserCheck(context, request)) return null;//如果该请求无需用户，则
    	if(isNoLoign()) {//快速切换用户需要关闭登录验证，防止生产环境安全问题
	        HttpCookie httpCookie = request.getCookies().getFirst(CommonConstant.AUTHORIZATION);
	        if(null != httpCookie) return httpCookie.getValue();
    	}
    	if(token==null) {
            context.setDoNext(false);
    		return null;
    	}
        try {
            boolean isLogin = true;
            String userId = clientUserUtil.getCasServerUserId(request,token);
            if (!StringUtils.isEmpty(userId)) {
                context.setUserToken(token);
            } else {
                unLogin(context);
            }
            return userId;
        } catch (Exception e) {
        	logger.error("Invoke SSO Exception :{}", e.getMessage());
            context.setDoNext(false);
        }
        return null;
    }

    /**
     * 鉴权
     * @param context 网关上下文
     * @param userId  因在网关中使用feign调用 网关无法传递cookie,门户FeignRequestInterceptor也无法拦截到因此或失去Cookie，门户无法获取到cookie 故使用userId传参
     * @param request
     */
    private void checkPermission(GatewayContext context, String userId,ServerHttpRequest request ) {
        //Url资源权限验证
        if(userId==null) {
            context.setDoNext(false);
        }
        try{
            String path = request.getURI().getPath();
            String method = request.getMethod().toString();
            //path 去除网关路由path
            path = path.substring(1,path.length());
            int index = path.indexOf("/");
            if(!noPermission(request)) {
                if (index > -1) {
                    path = path.substring(index + 1, path.length());
                    //调用门户资源权限校验
                    AjaxJson ajaxJson = permissionService.checkUrlPermission(userId, path, method);
                    if (ajaxJson.getCode().equals(LoitStatusCode.SUCESS.statusCode)) {
                        Boolean flag = (Boolean) ajaxJson.getBody().get("flag");
                        if (!flag) {
                            context.setDoNext(false);
                        }
                    }
                } else {
                    context.setDoNext(false);
                }
            }

        }catch (Exception e){
            logger.error("网关调用门户资源权限验证异常:{}",e);
            context.setDoNext(false);
        }
    }

	/**
     * 901状态码 用户未登录
     * @param response
     */
    private Mono<Void> get901Code(ServerHttpResponse response,ServerHttpRequest request){
    	URI uri = request.getURI();
    	String hostIp = uri.toString().substring(0, uri.toString().indexOf(uri.getPath()));
    	//TODO 通过获取uri的地址来跳转到登录页面可能会有问题，通过nginx转发gateway集群时时需要注意一下这个问题
        ClientAjaxJson ajaxJson = ClientAjaxJson.returnExceptionInfo(ClientLoitStatusCode.CAS_USER_TOKEN_FAIL);
        ajaxJson.put("loginUrl", hostIp +getPortalLogoutName() + "/cas/logout");
        byte[] bits = ajaxJson.getJsonStr().getBytes(StandardCharsets.UTF_8);
        DataBuffer buffer = response.bufferFactory().wrap(bits);
        response.getHeaders().setContentType(MediaType.APPLICATION_JSON_UTF8);
        return response.writeWith(Flux.just(buffer));
    }
	/**
     * 903状态码 无权限访问该资源
     */
    private Mono<Void> get903Code(ServerHttpResponse response){
        ClientAjaxJson ajaxJson = ClientAjaxJson.returnExceptionInfo(ClientLoitStatusCode.CAS_AUTHENTICATION_FAIL);
        byte[] bits = ajaxJson.getJsonStr().getBytes(StandardCharsets.UTF_8);
        DataBuffer buffer = response.bufferFactory().wrap(bits);
        response.getHeaders().setContentType(MediaType.APPLICATION_JSON_UTF8);
        return response.writeWith(Flux.just(buffer));
    }

    /**
     * + "?returnUrl=" + request.getURI()
     */
    private void unLogin(GatewayContext context) {
        context.setDoNext(false);
    }

    /**
     * 放行的url请求检查
     * @param context
     * @param exchange
     * @return
     */
    private boolean whiteListCheck(GatewayContext context, ServerWebExchange exchange) {
        String url = exchange.getRequest().getURI().getPath();
        for (String excludeUrl : authSkipUrlsProperties.getApiUrls()) {
			if(this.ant.match(excludeUrl, url)) {
				return true;
			}
		}
        return false;
    }

    /**
     * 不需要用户的请求
     * @param context
     * @return
     */
    private boolean noUserCheck(GatewayContext context, ServerHttpRequest request) {
        String url = request.getURI().getPath();
        //不需要用户信息的服务过滤
        for (String serverPath : authSkipUrlsProperties.getNoUserServers()) {
			if(this.ant.match(serverPath, url)) {
				return true;
			}
		}
    	String uri = url.substring(1);
    	int count = uri.indexOf("/");
    	//不需要用户信息的相对请求过滤
    	if(count!=-1) {
    		url = uri.substring(uri.indexOf("/"));
	        for (String excludeUri : authSkipUrlsProperties.getNoUserUri()) {
				if(this.ant.match(excludeUri, url)) {
					return true;
				}
			}
    	}


        return false;
    }


    /**
     * 无需资源权限校验
     * @param request
     * @return
     */
    private boolean noPermission(ServerHttpRequest request) {
        String url = request.getURI().getPath();
        String uri = url.substring(1);
        int count = uri.indexOf("/");
        //不需要资源权限校验的相对请求过滤
        if(count!=-1) {
            url = uri.substring(uri.indexOf("/"));
            for (String excludeUri : authSkipUrlsProperties.getCommonUri()) {
                if(this.ant.match(excludeUri, url)) {
                    return true;
                }
            }
        }
        return false;
    }
/*
    private boolean blackServersCheck(GatewayContext context, ServerWebExchange exchange) {
        String instanceId = exchange.getRequest().getURI().getPath().substring(1, exchange.getRequest().getURI().getPath().indexOf('/', 1));
        if (!CollectionUtils.isEmpty(authSkipUrlsProperties.getDataWorksServers())) {
        	boolean black = authSkipUrlsProperties.getServerPatterns().stream()
                    .map(pattern -> pattern.matcher(instanceId))
                    .anyMatch(Matcher::find);
            if (black) {
                context.setBlack(true);
                return true;
            }
        }
        return false;
    }
*/


    @Override
    public int getOrder() {
        return Ordered.HIGHEST_PRECEDENCE;
    }
}